// Tandy TRS-80 Clone
// PS/2 keyboard interface
// Glen Kleinschmidt
// www.glensstuff.com
// Jan 2020
// V1.0

// At the time of writing (Jan 2020) this is the highest level of development of this firmware.
// Note that I'm a hardware guy, not(!) a programmer and this code thus far was written with simplicity and
// expediency as the overruling factor (too many projects, not enuf time!). However all required
// functionality is implemented. The lack of finesse is in the absense of a watchdog timer and the primitive
// routines for receiving and writing data to and from the PS/2 keyboard with the current lack of parity checking.
// However, thus far, having spent many hours typing data into my TRS-80 and having literally typed thousands
// and thousands of characters without a single evident communications glitch, I'm happy to release this firmware
// as is for now. When I get around to it, it will eventually be written properly/updated and posted to my site.
// So far the only operational "difficulty" which I have encountered is that the serial routine can be 
// invoked to lock up (until power-cycled) if you madly bang the flat palms of your hands up and down on the
// keyboard to simultaneously press as many keys as possible.


#include <16f876A.h>
#use delay(clock=10000000)
//#use fast_io(A)
#use fast_io(A)
#use standard_io(B)
#use standard_io(C)
#FUSES HS,NOWDT,PUT,NOPROTECT,NOLVP,NOCPD,NOWRT,BROWNOUT


int1     command_parity     =  0;
int      command            =  0;
int      key_code           =  0;
int1     start_bit          =  0;
int1     parity_bit         =  0;   
int1     stop_bit           =  0;
int1     release            =  0;           
int      x;
int1     key_register[52]   = {0};
                           
             
void set_key_register()
{

   //                                       0   1   2   3   4   5   6   7   8   9   
   int const ps2_trs80_conv [128]   =     { 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, // 9
                                            99, 99, 99, 99,  0, 99, 99, 99, 51, 99, // 19
                                            99, 17, 28, 99, 99, 99, 26, 19,  1, 23, // 29
                                            29, 99, 99,  3, 24,  4,  5, 31, 30, 99, // 39
                                            99, 50, 22,  6, 20, 18, 32, 99, 99, 14, // 49
                                             2,  8,  7, 25, 33, 99, 99, 99, 13, 10, // 59
                                            21, 34, 35, 99, 99, 39, 11,  9, 15, 27, // 69
                                            36, 99, 99, 41, 42, 12, 37, 16, 40, 99, // 79
                                            99, 99, 99, 99, 99, 38, 99, 99, 99, 51, // 89
                                            43, 99, 99, 99, 99, 99, 99, 99, 99, 99, // 99
                                            99, 99, 48, 99, 99, 99, 99, 48, 44, 99, // 109
                                            99, 99, 99, 99, 47, 99, 49, 46, 45, 99, // 119
                                            99, 99, 99, 99, 99, 99, 99, 99};        // 127   
                                            
   
   key_code =  ps2_trs80_conv[key_code];
            
   
   if (release==0)
   {
      key_register[key_code]=1;
   }
   else
   {
      key_register[key_code]=0;
   }      
}

                                                                                                 
void key_code_to_port()
{
   
   key_code=0;  
   for (x=0; x<=7; x++)
   {
      if (key_register[x] == 1)
      bit_set(key_code,x);
   }
   output_C(key_code);
   output_high(pin_B0);
   output_low(pin_B0);
   
   key_code=0;  
   for (x=0; x<=7; x++)
   {
      if (key_register[x+8] == 1)
      bit_set(key_code,x);
   }
   output_C(key_code);
   output_high(pin_B1);
   output_low(pin_B1);
  
   key_code=0;  
   for (x=0; x<=7; x++)
   {
      if (key_register[x+16] == 1)
      bit_set(key_code,x);
   }
   output_C(key_code);
   output_high(pin_B2);
   output_low(pin_B2);
 
   key_code=0;  
   for (x=0; x<=7; x++)
   {
      if (key_register[x+24] == 1)
      bit_set(key_code,x);
   }
   output_C(key_code);
   output_high(pin_B3);
   output_low(pin_B3);  
   
   key_code=0;  
   for (x=0; x<=7; x++)
   {
      if (key_register[x+32] == 1)
      bit_set(key_code,x);
   }
   output_C(key_code);
   output_high(pin_B4);
   output_low(pin_B4);  
   
   key_code=0;  
   for (x=0; x<=7; x++)
   {
      if (key_register[x+40] == 1)
      bit_set(key_code,x);
   }
   output_C(key_code);
   output_high(pin_B5);
   output_low(pin_B5);   
   
   key_code=0;  
   for (x=0; x<=3; x++)
   {
      if (key_register[x+48] == 1)
      bit_set(key_code,x);
   }
   output_C(key_code);
   output_high(pin_A5);
   output_low(pin_A5);   

}


void serial_read()
{

   // This routine receives data (key make and break codes) from the PS/2 keyboard 
   // A0 = Keyboard serial clock input
   // A1 = Keyboard serial data input
   
   key_code=0;
   start_bit=0;
   parity_bit=0;
   stop_bit=0;
  
   while(input(pin_A0));         // Start bit      
   
   if (input(pin_A1))
   start_bit=1;
     
   while(!input(pin_A0));
       
   while(input(pin_A0));         // Bit 0
      
   if (input(pin_A1))
   bit_set(key_code,0);
   
   while(!input(pin_A0));
        
   while(input(pin_A0));         // Bit 1
   
   if (input(pin_A1))
   bit_set(key_code,1);
      
   while(!input(pin_A0));
        
   while(input(pin_A0));         // Bit 2
  
   if (input(pin_A1))
   bit_set(key_code,2);
   
   while(!input(pin_A0));
         
   while(input(pin_A0));         // Bit 3
   
   if (input(pin_A1))
   bit_set(key_code,3);   
   
   while(!input(pin_A0));
         
   while(input(pin_A0));         // Bit 4
   
   if (input(pin_A1))
   bit_set(key_code,4);  
   
   while(!input(pin_A0));
       
   while(input(pin_A0));         // Bit 5
   
   if (input(pin_A1))
   bit_set(key_code,5);  
   
   while(!input(pin_A0));
       
   while(input(pin_A0));         // Bit 6
   
   if (input(pin_A1))
   bit_set(key_code,6);
         
   while(!input(pin_A0));
     
   while(input(pin_A0));         // Bit 7
   
   if (input(pin_A1))
   bit_set(key_code,7);
   
   while(!input(pin_A0));
         
   while(input(pin_A0));         // Parity bit
   
   if (input(pin_A1))
   parity_bit=1;
   
   while(!input(pin_A0));
        
   while(input(pin_A0));         // Stop bit
   
   if (input(pin_A1))
   stop_bit=1;
   
   while(!input(pin_A0));
            
}


void serial_write()
{

   // This routine sends data to the PS/2 keyboard 
   // A0 = Keyboard serial clock input
   // A1 = Keyboard serial data input

   set_tris_A(254);              // A0 (clock) = output; 11111110
   output_low(pin_A0);           // Clock line low
   delay_us(100);
   set_tris_A(252);              // A0 (clock) and A1 (data) = output; 11111100
   output_low(pin_A1);           // Data line low
   set_tris_A(253);              // A1 (data) = output, A0 (clock) = input (release clock); 11111101   
   delay_us(10);
   
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,0));

   while(!input(pin_A0));
  
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,1));

   while(!input(pin_A0));
    
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,2));

   while(!input(pin_A0));
     
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,3));

   while(!input(pin_A0));
       
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,4));

   while(!input(pin_A0));
     
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,5));

   while(!input(pin_A0));
     
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,6));

   while(!input(pin_A0));
     
   while(input(pin_A0));
   
   output_bit(pin_A1, bit_test(command,7));

   while(!input(pin_A0));

   while(input(pin_A0));                        // Parity bit
   
   output_bit(pin_A1, command_parity);

   while(!input(pin_A0));
  
   while(input(pin_A0));
   
   set_tris_A(223);                             // Release the data line (A5 = output; 11011111)

}


void initialise_registers()
{

   output_C(0);
   
   output_high(pin_B0);
   output_low(pin_B0);
   
   output_high(pin_B1);
   output_low(pin_B1);
   
   output_high(pin_B2);
   output_low(pin_B2);
   
   output_high(pin_B3);
   output_low(pin_B3);
   
   output_high(pin_B4);
   output_low(pin_B4);
   
   output_high(pin_B5);
   output_low(pin_B5);
   
   output_high(pin_B6);
   output_low(pin_B6);
   
   output_high(pin_A5);
   output_low(pin_A5);

}


void initialise_keyboard()
{      

   // This routing initialises the PS/2 keyboard with fixed constants to ensure correct operation. The keyboard 
   // must be plugged in at power-on.
   
   command           =  255;                    // Reset keyboard
   command_parity    =  1; 
   serial_write();
         
   delay_ms(1000);
   
   command           =  243;                    // Set typmatic delay
   command_parity    =  1;
   serial_write();
   
   delay_ms(100);
   
   command           =  127;                    // 1 - 2cps 
   command_parity    =  0;
   serial_write();
   
   delay_ms(100);

}


void ps2()
{

   // This routine acts upon data received from the PS/2 Keyboard. 
   // See the associated doccumentation for a printable graphical keyboard
   // chart/key-map to aid the typist.

 
   Serial_read();          
   
           
   if(key_code==0xF0)                        // Detect release code 
   {                          
      serial_read();                         // Receive code for released key
      release=1;                        
   }
   else
   {
      release=0;
   }
   
     
   if(key_code==0xE0)                        // Detect Extended key code
   {
   
      serial_read();                         // Receive key code for extended key
   
      if(key_code==0xF0)                     // Detect Extended key release code
      {
         serial_read();                      // Receive code for released extended key
         release=1;
      }
      else
      {
         release=0;
      }
   }
   
   
   set_key_register();                       // Update the key status register
   key_code_to_port();                       // Send key_code to matrix encoder

}


void MAIN()
{
   
   set_tris_A(223);                          // A5 = output; 11011111
   
   delay_ms(100);
   initialise_registers();
   initialise_keyboard();
   
   loop:

   ps2();

   goto loop;
   
}

